--##****************************************************
--##**	ZONE CONTROL FOR SUPREME COMMANDER    **
--##**			        BY SAYA                                          **
--##**			       V 1.0 BETA                                        **
--##****************************************************

-- MAP BY ANGRY ZEALOT

--scenario utilities
local ScenarioUtils = import('/lua/sim/ScenarioUtilities.lua')
local ScenarioFramework = import('/lua/ScenarioFramework.lua')
local Utilities = import('/lua/utilities.lua')

--interface utilities
local UIUtil = import('/lua/ui/uiutil.lua')
local gameParent = import('/lua/ui/game/gamemain.lua').GetGameParent()

local executing = false
local beatTime = 5

--TRIGGER TABLES
local bunkerDeath = {}
local unitDeath = {}
	
--PLAYERS, CONTAINS USEFUL INFO FOR THE SCRIPT FOR EACH PLAYER
local playersDATA = {
	['player1'] = {
		name = 'player1', 
		kills = 0, 
		bunkerType = 'ueb2301',
		zonecontrol = 1, 
		attack = 0, 
		defence = 0, 
		upgrades = 1, 
		level = 0,
		money = 0,
		infoUnit = false,
		attackUnit = false,
		defenceUnit = false,
		upgrader = false,
		startLocation = {1,4},
		baseTemplate = {info = {8,42}, attack = {8,20}, defence = {30,42}, upgrader = {20,30}},
	},
	['player2'] = {
		name = 'player2', 
		kills = 0, 
		bunkerType = 'ueb2301',
		zonecontrol = 1, 
		attack = 0, 
		defence = 0, 
		upgrades = 1, 
		level = 0,
		money = 0,
		infoUnit = false,
		attackUnit = false,
		defenceUnit = false,
		upgrader = false,
		startLocation = {2,1},
		baseTemplate = {info = {8,8}, attack = {8,30}, defence = {30,8}, upgrader = {20,20}},
	},
	['player3'] = {
		name = 'player3',
		kills = 0, 
		bunkerType = 'uab2301',
		zonecontrol = 1, 
		attack = 0, 
		defence = 0, 
		upgrades = 1, 
		level = 0,
		money = 0,
		infoUnit = false,
		attackUnit = false,
		defenceUnit = false,
		upgrader = false,
		startLocation = {5,2},
		baseTemplate = {info = {42,8}, attack = {20,8}, defence = {42,30}, upgrader = {30,20}},
	},
	['player4'] = {
		name = 'player4', 
		kills = 0, 
		bunkerType = 'uab2301',
		zonecontrol = 1, 
		attack = 0, 
		defence = 0, 
		upgrades = 1, 
		level = 0,
		money = 0,
		infoUnit = false,
		attackUnit = false,
		defenceUnit = false,
		upgrader = false,
		startLocation = {4,5},
		baseTemplate = {info = {42,42}, attack = {42,20}, defence = {20,42}, upgrader = {30,30}},
	},
	['neutral'] = {
		name = 'neutral', 
		kills = 0, 
		bunkerType = 'urb2301',
		zonecontrol = 0, 
		attack = 0, 
		defence = 0, 
		upgrades = 1, 
		level = 0,
		money = 0,
		infoUnit = false,
		attackUnit = false,
		defenceUnit = false,
		upgrader = false,
	},
}

local players = {}

local defeatcount = 0
local gameEnd = false
local beat = 0



local armyTable = {
	 ['ARMY_1'] = 'player1',
	 ['ARMY_2'] = 'player2',
	 ['ARMY_3'] = 'player3',
	 ['ARMY_4'] = 'player4',
	 ['ARMY_9'] = 'neutral',
}


local upgradeList = {
	['LEVEL 1'] = { level = 1, kills = 50, upgrade = 2, description = 'LEVEL 1: TANK 400 HP'},   
	['LEVEL 2'] = { level = 2, kills = 200, upgrade = 3, description = 'LEVEL 2: PILLAR 600 HP'},
	['LEVEL 3'] = { level = 3, kills = 400, upgrade = 7, description = 'LEVEL 3: ASSAULT BOT 1400 HP'},
	['LEVEL 4'] = { level = 4, kills = 600, upgrade = 7, description = ''},
	['LEVEL 5'] = { level = 5, kills = 900, upgrade = 7, description = ''},
}

local unitUpgradeList = {
	['LEVEL 0'] = 'UEL0106',--level 0: mech marine
	['LEVEL 1'] = 'UEL0201',--level 1: mantis
	['LEVEL 2'] = 'UEL0202',--level 2: pillar
	['LEVEL 3'] = 'UEL0303',--level 3: assault bot
	['LEVEL 4'] = 'UEL0303',--level 4: assault bot
	['LEVEL 5'] = 'UEL0303',--level 5: assault bot
}



--create areas on the map
local zones = {}
local xzonesize = 50
local yzonesize = 50
local xmapsize = 256
local ymapsize = 256
local x = 0
local y = 0

--number of zones on the X axis
local imax = math.floor(xmapsize / xzonesize)
--number of zones on the Y axis
local jmax = math.floor(ymapsize / yzonesize)







--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~UTILITY FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

--given a player returns an army
toArmy = function(player)
	local army = false
	for i,v in armyTable do
		if v == player.name then
			army = i
		end
	end
	return army
end


--given a player returns a proper username
username = function(player)
	return GetArmyBrain(toArmy(player)).Nickname
end

--given a player returns the number of kills
armyKills = function(player)
	local kills = 0
	if player.name != 'neutral' then 
		local army = toArmy(player)
		kills = GetArmyBrain(army):GetArmyStat('Enemies_Killed',0.0).Value 
	end
	return kills	
end

--print a text scn way :p
scnPrint = function(text,player)
	PrintText(text,20,GetArmyBrain(toArmy(player)).color,beatTime,'center') 
end

--given a unit returns the army
scnArmy = function(unit)
	local armyIndex = unit:GetArmy()
	local armycount = 0
	local army = ''
	if players['player1'] then
		armycount = armycount + 1
		if armyIndex == armycount then
			army = 'ARMY_1'
		end
	end
	if players['player2'] then
		armycount = armycount + 1
		if armyIndex == armycount then
			army = 'ARMY_2'
		end
	end
	if players['player3'] then
		armycount = armycount + 1
		if armyIndex == armycount then
			army = 'ARMY_3'
		end
	end
	if players['player4'] then
		armycount = armycount + 1
		if armyIndex == armycount then
			army = 'ARMY_4'
		end
	end
	if players['neutral'] then
		armycount = armycount + 1
		if armyIndex == armycount then
			army = 'ARMY_9'
		end
	end	
	
	return army
end

--get the number of players
playerCount = function()
	local i = 0
	for index, player in players do
		i = i + 1
	end
	return i
end








--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~MAIN SCRIPT FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

--populating
function OnPopulate()
	InitializeArmies()
    ScenarioFramework.SetPlayableArea('Playfield', false)
end


--start of map script
function OnStart(self)
	--initialise scenario
	ScenarioFramework.SetPlayableArea('Playfield', false)
	Utilities.UserConRequest("ui_ForceLifbarsOnEnemy")--show enemy life bars
	initialiseZones()
	
	--initialise player zone control
	for index, player in players do
		if player.startLocation != nil then
			changeOwner(zones[player.startLocation[1]][player.startLocation[2]],player.name)
		end
	end
	
	--spawn start bunkers
	spawnBunkers()
	
	--spawn start info units
	spawnUtilityUnits()
	
	--start main thread
	local mainThread = ForkThread(function()
		while true do
			--secutiry for slow computers
			while executing == true do
			end
			--call main thread beat
			mainThreadBeat()
			--wait time between each beat
			WaitSeconds(beatTime)
			--increment beat count
			--LOG('TICKS: ' .. beat)
			beat = beat + 1
		end
    end)
end

--MAIN SCENARIO THREAD
--do map script actions every TICK
mainThreadBeat = function()
	executing = true
	checkUpgrades()
	checkZones()
	spawnUnits()
	checkVictoryConditions()
	executing = false
end







--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~MAIN THREAD FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

--check for end game conditions
checkVictoryConditions = function()
	--check for defeat
	for index, player in players do
		--LOG(player.zonecontrol)
		if player.zonecontrol == 0 and gameEnd == false then
			defeatcount = defeatcount + 1
			player.zonecontrol = -1
			GetArmyBrain(toArmy(player)):OnDefeat() 
		end
	end
	--check for victory
	if defeatcount == (playerCount()-1) and gameEnd == false then
		for index, player in players do 
			if player.zonecontrol >= 1 then
				GetArmyBrain(toArmy(player)):OnVictory()
			end
		end
		gameEnd = true
	end
end





--CHECK IF UPGRADES ARE NECESSARY, APPLY THEM
checkUpgrades = function()
	--WARN("TEST: " .. players['player1'].level)
	--APPLY UPGRADES
	for i,v in upgradeList do
		for i2,v2 in players do
			--WARN(v2.name .. " kills: " .. armyKills(v2))
			if armyKills(v2) >= v.kills then
				--WARN("TRYING TO UPGRADE TO LEVEL ".. v.level)
				upgrade(v2,i)
			end
		end
	end
	checkUpgradeRequests()
end



--SPAWN UNITS ON ALL ZONES DEPENDING ON OWNERSHIP
spawnUnits = function()
	local i = 1
	while i <= imax do
		local j = 1
		while j <= jmax do
			if (i > 1 and i < 5) or (j > 1 and j<5) then
			spawnUnit(zones[i][j])
			end
			j = j + 1
		end
		i = i + 1
	end	
end


--SPAWN BUNKERS AT START
spawnBunkers = function()
	local i = 1
	while i <= imax do
		local j = 1
		while j <= jmax do
			if (i > 1 and i < 5) or (j > 1 and j<5) then
			spawnBunker(zones[i][j])
			end
			j = j + 1
		end
		i = i + 1
	end	
end

--CHECK OWNERSHIP ON ALL ZONES
checkZones = function()
	local i = 1
	while i <= imax do
		local j = 1
		while j <= jmax do
			if (i > 1 and i < 5) or (j > 1 and j<5) then
			checkZone(zones[i][j])
			end
			j = j + 1
		end
		i = i + 1
	end	
end

--make all zones
initialiseZones = function()
	local i = 1
	while i <= imax do		
		local j = 1
		zones[i] = {}
		bunkerDeath[i] = {}
		while j <= jmax do
			if (i > 1 and i < 5) or (j > 1 and j<5) then
				zones[i][j] = createZone(i,j)
			end
			j = j + 1
		end
		i = i + 1

	end
end












--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ZONE MANAGING FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
--zone constructor
createZone = function(i,j)
	--create zone at (i, j) coordonates
	local zone =  {owner = 'neutral', x = i, y = j, Xsize = xzonesize, Ysize = yzonesize, hasBunker = false}
	
	--create zone rectangle
	zone.rectangle = {}
	zone.rectangle.x0 = (zone.x - 1) * zone.Xsize + 1 
	zone.rectangle.y0 = (zone.y - 1) * zone.Ysize + 5
	zone.rectangle.x1 = zone.rectangle.x0 + zone.Xsize
	zone.rectangle.y1 = zone.rectangle.y0 + zone.Ysize
	
	--generate ,bunker death trigger functions
	bunkerDeath[i][j] = function()		
		--acknowledge bunker destruction and refresh zone count
		zone.hasBunker = false
		removeZone(zone.owner)		
	end

	return zone
end

--given a zone, returns the position of its center
zoneCenter = function(zone)
	local center = {}
	center.x = zone.x * zone.Xsize - ( zone.Xsize / 2) + 1
	center.y = zone.y * zone.Ysize - ( zone.Ysize / 2) + 5
	return center
end


--change the owner of a zone
changeOwner = function(zone,newowner)
	zone.owner = newowner
end




--CHECK ZONE OWNERSHIP, CHANGE IT IF NECESSARY
checkZone = function(zone)
	--IF ZONE CONTAINS A BUNKER DO NOT CHECK: DONE
	--IF ZONE CONTAINS ONLY UNITS FROM 1 PLAYER AND PLAYER != OWNER THEN OWNER = PLAYER	
	--CREATE A BUNKER FOR THE NEW OWNER: DONE	
	
	--if zone does not have a bunker
	if zone.hasBunker == false then --and zone.owner == 'neutral' then
		--find new owner
		local ownerfound = refreshOwner(zone)
		--spawn bunker on zone
		if zone.owner != 'neutral' and ownerfound == true then
		spawnBunker(zone)		
		--add zone to owner
		addZone(zone.owner)		
		end

	end	

end

--FIND NEW OWNER FOR ZONE
refreshOwner = function(zone)
	local ownerfound = false
	
	--LOG(repr(zone.rectangle))
	
	local ally = toArmy(players[zone.owner])	
	local count = { ['ARMY_1'] = 0, ['ARMY_2'] = 0, ['ARMY_3'] = 0, ['ARMY_4'] = 0, ['ARMY_9'] = 0}
	local enemyCount = 0
	
	--get units in zone
	local unitsInZone = GetUnitsInRect(zone.rectangle)
	

	
	--get army counts in zone
	if unitsInZone then
		for k,unit in unitsInZone do
			if not unit:IsDead()  then
				local army = ''
				if unit:GetArmy() == playerCount() then
					army = 'ARMY_9'
				else
					army = scnArmy(unit)
				end
				
				count[army] = count[army] + 1
				if army != ally then 
					enemyCount = enemyCount + 1
				end
			end
		end
	end
	
	--determine new owner depending on army count
	--zone owner retakes ownership
	if count[ally] >= 10 and enemyCount == 0 then
		ownerfound = true
	--attacker captures zones
	elseif count[ally] <= 5 then
		local maxcountArmy = false
		local maxvalue = 0
		local draw = false
		
		--find army with most troops on zone
		for army,value in count do 
			if maxcountArmy == false then
				maxcountArmy = army
			end
			if value > maxvalue then
				maxvalue = value
				maxcountArmy = army
			elseif value == maxvalue and value != 0 then
				draw = true
			end
		end
		
		--if the army with more troops has more than 5 troops then capture the zone
		if count[maxcountArmy] >= 5 and draw == false then
			zone.owner = armyTable[maxcountArmy]
			ownerfound = true
		end
		
		
	end
	
	return ownerfound
end


--add zone to count of player
addZone = function(playername)
	players[playername].zonecontrol = players[playername].zonecontrol + 1
	--instigator:GetAIBrain():GiveResource('MASS', mamount) 
	--armyTable:GetAIBrain():GiveResource('ENERGY', eamount) 
	--GetArmyBrain(toArmy(players[playername])):GiveResource('MASS', 1) 
	--LOG("" .. players[player].name .."ZONE CONTROL: " .. players[player].zonecontrol)
end

--remove zone from count of player
removeZone = function(playername)
	players[playername].zonecontrol = players[playername].zonecontrol - 1
	--GetArmyBrain(toArmy(players[playername])):TakeResource('MASS', 1) 
	--self:GetAIBrain():TakeResource('Energy', self:GetBlueprint().Economy.StorageEnergy) 
	--LOG("" .. players[player].name .."ZONE CONTROL: " .. players[player].zonecontrol)
end










--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~UNIT SPAWNING FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


--spawns a bunker on a zone
spawnBunker = function(zone)
	--SPAWN UNIT ON ZONE DEPENDING ON PLAYER OWNERSHIP
	--SPAWN NO UNIT IF ZONE IS NEUTRAL
	local x = zoneCenter(zone).x
	local y = zoneCenter(zone).y
	local player = players[zone.owner]
	local army = toArmy(player)
	
	zone.bunker = false
	
	--spawn bunker if not population limit
	if GetArmyUnitCostTotal(GetArmyBrain(army):GetArmyIndex()) < GetArmyUnitCap(GetArmyBrain(army):GetArmyIndex()) then		
		zone.bunker = CreateUnitHPR( player.bunkerType, toArmy(player), x,y,y, 0,0,0) 
	end
	
	--if bunker exists apply balance changes and upgrades
	if zone.bunker != false then
		zone.hasBunker = true
		
		if zone.owner == 'neutral' then 
			zone.bunker:SetMaxHealth(600)
			zone.bunker:SetHealth(zone.bunker,600)
		else
			zone.bunker:SetMaxHealth(5000 + 1000*player.defence)
			zone.bunker:SetHealth(nil,5000 + 1000*player.defence)
			zone.bunker:SetRegenRate(10*player.defence)
			
			for i = 1, zone.bunker:GetWeaponCount() do
				zone.bunker:GetWeapon(i):ChangeRateOfFire(2 + player.defence/2)
				zone.bunker:GetWeapon(i):AddDamageMod(50*player.defence)
			end
			
			zone.bunker:SetCustomName(username(player) .. ', level ' .. player.defence)
			
			addBunkerKillCount(zone.bunker)
		end
		zone.bunker:GetWeapon(1):ChangeMaxRadius(30)
		ScenarioFramework.CreateUnitDeathTrigger(bunkerDeath[zone.x][zone.y],zone.bunker)
	end

end

--ADD A KILL COUNT TO A BUNKER
--defender gets more money than attacker, a kill made by a bunker gives 10 instead of 5 for a regular kill
addBunkerKillCount = function(unit)
	--kill count
	unit.OldOnKilledUnit = unit.OnKilledUnit
	unit.OnKilledUnit = function(self, unitKilled)
		local player = players[armyTable[scnArmy(self)]]
		--if the unit did not commit suicide		
		if self:GetArmy() != unitKilled:GetArmy() then 
			player.kills = player.kills + 1
			player.money = player.money + 10
			--update display
			updateInfo(player)			
			--GetArmyBrain(toArmy(player)):GiveResource('ENERGY', 10000) 		
		end
		unit.OldOnKilledUnit(self, unitKilled)	
	end		
end


--SPAWN UNIT ON ZONE
spawnUnit = function(zone)
	--SPAWN UNIT ON ZONE DEPENDING ON PLAYER OWNERSHIP
	--SPAWN NO UNIT IF ZONE IS NEUTRAL
	local x = zoneCenter(zone).x
	local y = zoneCenter(zone).y
	local unit = false	
	local player = players[zone.owner]
	local army = toArmy(player)
	local unitType = unitUpgradeList['LEVEL ' .. player.level]
	
	local xside = math.random(0,1)
	local yside = math.random(0,1)
	local xoff = 0
	local yoff = 0
	
	if zone.hasBunker == false then
	elseif zone.owner != 'neutral' then
		--check unit cap to avoid bug		
		if GetArmyUnitCostTotal(GetArmyBrain(army):GetArmyIndex()) < GetArmyUnitCap(GetArmyBrain(army):GetArmyIndex()) then
			--adjust spawn rate depending on level
			if player.level <= 1 then --low level units spawn every tick
					--unit = CreateUnitHPR( unitType, army, x,25,y+3, 0,0,0) 		
					unit = CreateUnitHPR( unitType, army, x+(-1+xside*2),25,y+(-1+yside*2), 0,0,0)	
			elseif player.level == 2 then -- pillars spawn 1 time out of 2
				if math.mod(beat,2) == 0 then
					--unit = CreateUnitHPR( unitType, army, x,25,y+3, 0,0,0) 	
					unit = CreateUnitHPR( unitType, army, x+(-1+xside*2),25,y+(-1+yside*2), 0,0,0)
				end
			elseif player.level >=3 then -- siege bots spawn 1 time out of 3
				if math.mod(beat,3) == 0 then
					--unit = CreateUnitHPR( unitType, army, x,25,y+3, 0,0,0) 	
					unit = CreateUnitHPR( unitType, army, x+(-1+xside*2),25,y+(-1+yside*2), 0,0,0)
				end
			end
		end
	end
	
	--if unit exists apply unit balance changes and upgrades
	if unit != false then
		addKillCount(unit)
		
		unit:DestroyShield()
		unit:SetMaxHealth(200*upgrades(zone)+50*player.attack)
		unit:SetHealth(nil,200*upgrades(zone)+50*player.attack)
		
		--thanks to UArchitect
		--disable wreckage as way too many of the stuff appears..decals are enough, (wouldnt be so bad if i could make them slowly sink into the ground after a while)
		unit:GetBlueprint().Wreckage.WreckageLayers.Land=false
		
        for i = 1, unit:GetWeaponCount() do
            unit:GetWeapon(i):AddDamageMod(5*player.attack)
        end
		
		unit:SetSpeedMult(2)
		unit:SetAccMult(3)       
        unit:SetTurnMult(3)
		--ScenarioFramework.CreateUnitDeathTrigger(unitDeath[zone.owner],unit)		
		--move unit after being created
		
		--thanks to UArchitect
		--add randomness stuff to the move order
		xoff = -8 + xside*16
		yoff = -8 + yside*16
		
		xoff = xoff + -2 + math.random(0,1)*4
		yoff = yoff + -2 + math.random(0,1)*4
		
		IssueMove( {unit}, {x+xoff,25,y+yoff} )
		--IssueMove( {unit}, {x,25,y+8} )
		
	end
end

--ADD A KILL COUNT TO A UNIT
addKillCount = function(unit)
	--kill count
	unit.OldOnKilledUnit = unit.OnKilledUnit
	unit.OnKilledUnit = function(self, unitKilled)
		local player = players[armyTable[scnArmy(self)]]
		--if the unit did not commit suicide		
		if self:GetArmy() != unitKilled:GetArmy() then 
			player.kills = player.kills + 1
			player.money = player.money + 5
			--update display
			updateInfo(player)			
			--GetArmyBrain(toArmy(player)):GiveResource('ENERGY', 10000) 		
		end
		unit.OldOnKilledUnit(self, unitKilled)	
	end		
end



--UNIT DEATH TRIGGERS
unitDeath['player1'] = function()
	--players['player1'].deaths = players['player1'].deaths + 1
	WARN('player1')
	WARN('index: ' .. GetArmyBrain(toArmy(players['player1'])):GetArmyIndex())
end

unitDeath['player2'] = function()
	--players['player2'].deaths = players['player2'].deaths + 1
	WARN('player2')
	WARN('index: ' .. GetArmyBrain(toArmy(players['player2'])):GetArmyIndex())
end

unitDeath['player3'] = function()
	--players['player3'].deaths = players['player3'].deaths + 1
	WARN('player3')
	WARN('index: ' .. GetArmyBrain(toArmy(players['player3'])):GetArmyIndex())
end

unitDeath['player4'] = function()
	--players['player4'].deaths = players['player4'].deaths + 1
	WARN('player4')
	WARN('index: ' .. GetArmyBrain(toArmy(players['player4'])):GetArmyIndex())
end









--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~UPGRADE SYSTEM FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

--check for attack or defence upgrade requests in area
checkUpgradeRequests = function()
	for index,player in players do
		--check for attack upgrade request
		local requestGroupA = {}
		if player.attackUnit != false then
			if GetArmyBrain(toArmy(player)):GetUnitsAroundPoint(categories.MOBILE,player.attackUnit:GetPosition(), 10,'Ally') != nil then
				requestGroupA = GetArmyBrain(toArmy(player)):GetUnitsAroundPoint(categories.MOBILE,player.attackUnit:GetPosition(), 10,'Ally')
				if table.getsize(requestGroupA) != 0 then
					respawnUpgrader(player,player.baseTemplate.upgrader.px,player.baseTemplate.upgrader.py)
					upgradeAttack(player)
				end
			end
		end
		
		--check for defence upgrade request
		local requestGroupB = {}
		if player.defenceUnit != false then
			if GetArmyBrain(toArmy(player)):GetUnitsAroundPoint(categories.MOBILE,player.defenceUnit:GetPosition(), 10,'Ally') != nil then
				requestGroupB = GetArmyBrain(toArmy(player)):GetUnitsAroundPoint(categories.MOBILE,player.defenceUnit:GetPosition(), 10,'Ally')
				if table.getsize(requestGroupB) != 0 then
					respawnUpgrader(player,player.baseTemplate.upgrader.px,player.baseTemplate.upgrader.py)
					upgradeDefence(player)
				end
			end
		end
	end
end

--return the number of upgrades of the owner of a zone
upgrades = function(zone)
	return players[zone.owner].upgrades
end

--upgrades a player to a given level
upgrade = function(player,levelindex)
	if upgradeList[levelindex].level > player.level then
		player.level = upgradeList[levelindex].level
		player.upgrades = upgradeList[levelindex].upgrade
		
		--level specific actions
		if upgradeList[levelindex].level == 4 then
			local startzone = zones[player.startLocation[1]][player.startLocation[2]]
			local x = zoneCenter(startzone).x 
			local y = zoneCenter(startzone).y
			local hero = CreateUnitHPR( 'ual0401', toArmy(player), x,25,y, 0,0,0) 
			hero:SetMaxHealth(250000)
		    hero:SetHealth(hero,250000)
			hero:SetCustomName('Galactic Colossus: ' .. username(player))
			scnPrint(username(player) .. ": Galactic Colossus Has Arrived",player)
		elseif upgradeList[levelindex].level == 5 then
			local startzone = zones[player.startLocation[1]][player.startLocation[2]]
			local x = zoneCenter(startzone).x 
			local y = zoneCenter(startzone).y
			local hero = CreateUnitHPR( 'URL0402', toArmy(player), x,25,y, 0,0,0) 
			hero:SetMaxHealth(500000)
		    hero:SetHealth(hero,500000)
			hero:SetCustomName('Lord of Monkey Lords: ' .. username(player))
			scnPrint(username(player) .. ": Lord of Monkey Lords Has Arrived",player)
		else		
			scnPrint(username(player) .. " has reached " .. upgradeList[levelindex].description,player)
		end
	end
end

--upgrade attack level trigger
--attack upgrade upgrades unit weapon damage
upgradeAttack = function(player)
	if player.money >= (player.attack + 1)*50 then		
		player.money = player.money - (player.attack + 1)*50
		player.attack = player.attack + 1
		updateInfo(player)
		scnPrint(username(player) .. ': upgrading attack to level '..player.attack,player)
	else
		scnPrint(username(player) .. ': not enough money for attack upgrade',player)
	end
end

--upgrade defence level trigger
--defence upgrade upgrades unit health as well as tower regen rate
upgradeDefence = function(player)
	if player.money >= (player.defence + 1)*50 then
		player.money = player.money - (player.defence + 1)*50
		player.defence = player.defence + 1
		updateInfo(player)
		scnPrint(username(player) .. ': upgrading defence to level '..player.defence,player)
		upgradeBunkers(player)
	else		
		scnPrint(username(player) .. ': not enough money for defence upgrade',player)
	end
end
	
--upgrades all alive bunkers for a player
upgradeBunkers = function(player)
	local i = 1
	while i <= imax do
		local j = 1
		while j <= jmax do
			if (i > 1 and i < 5) or (j > 1 and j<5) then
				local zone = zones[i][j]
				if zone.owner == player.name and zone.bunker != false then
					if not zone.bunker:IsDead() then
					zone.bunker:SetRegenRate(10*(player.defence+1))
					zone.bunker:SetMaxHealth(5000 + 1000*player.defence)
					zone.bunker:SetHealth(nil,5000 + 1000*player.defence)
					for i = 1, zone.bunker:GetWeaponCount() do
						zone.bunker:GetWeapon(i):ChangeRateOfFire(2 + player.defence/2)
						zone.bunker:GetWeapon(i):AddDamageMod(50*player.defence)
					end
					
					zone.bunker:SetCustomName(username(player) .. ', level ' .. player.defence)
					end
				end
			end
			j = j + 1
		end
		i = i + 1
	end	
end


--kill and respawn player upgrader  unit
respawnUpgrader = function(player,x,y)
	local xPos = 20
	local yPos = 20
	
	if x != nil then
		xPos = x
	end
	
	if y != nil then
		yPos = y
	end
	
	player.baseTemplate.upgrader.px = x
	player.baseTemplate.upgrader.py = y

	--destroy upgrader unit if necessary
	if player.upgrader != false then
		player.upgrader:SetCanBeKilled(true)
		player.upgrader:Kill()
		player.upgrader = false
	end	
	--spawn upgrader unit
	if player.upgrader == false then 
		player.upgrader = CreateUnitHPR( 'URL0107', toArmy(player), xPos,25,yPos, 0,0,0) 
		player.upgrader:SetAllWeaponsEnabled(false)	
		player.upgrader:SetCanBeKilled(false)
		player.upgrader:SetDoNotTarget(true)
		player.upgrader:SetMaxHealth(1)
		player.upgrader:SetHealth(nil,1)
		player.upgrader:SetRegenRate(1)
		player.upgrader:SetSpeedMult(2)
		player.upgrader:SetAccMult(10)       
        --player.upgrader:SetTurnMult(10)
		player.upgrader:SetCustomName(username(player) .. ': Upgrader')
	end
end

--spawn invincible units used to show stats, make upgrades etc
spawnUtilityUnits = function()
	if players['player1'] then
		spawnPlayerUtilityUnits(players['player1'],1,5)
	end
	if players['player2'] then
		spawnPlayerUtilityUnits(players['player2'],1,1)
	end
	if players['player3'] then
		spawnPlayerUtilityUnits(players['player3'],5,1)
	end
	if players['player4'] then
		spawnPlayerUtilityUnits(players['player4'],5,5)
	end
end

--spawn invincible units used to show stats, make upgrades etc for a player at a given zone position
spawnPlayerUtilityUnits = function(player,x,y)

	local positionX = 50*(x-1)
	local positionY = 50*(y-1)
	local bt = player.baseTemplate
	
	--info unit
	player.infoUnit = CreateUnitHPR( 'uab5202', toArmy(player), positionX+bt.info[1],25,positionY+bt.info[2], 0,0,0) 
	player.infoUnit:SetAllWeaponsEnabled(false)		
	player.infoUnit:SetUnSelectable(true)
	player.infoUnit:SetCanBeKilled(false)
	player.infoUnit:SetDoNotTarget(true)
	player.infoUnit:SetMaxHealth(1)
	player.infoUnit:SetHealth(nil,1)
	player.infoUnit:SetRegenRate(1)
	
	--attack upgrade
	player.attackUnit = CreateUnitHPR( 'urb5202', toArmy(player), positionX+bt.attack[1],25,positionY+bt.attack[2], 0,0,0) 
	player.attackUnit:SetAllWeaponsEnabled(false)		
	player.attackUnit:SetCanBeKilled(false)
	player.attackUnit:SetDoNotTarget(true)
	player.attackUnit:SetMaxHealth(1)
	player.attackUnit:SetHealth(nil,1)
	player.attackUnit:SetRegenRate(1)

	--defence upgrade
	player.defenceUnit = CreateUnitHPR( 'ueb5202', toArmy(player), positionX+bt.defence[1],25,positionY+bt.defence[2], 0,0,0) 
	player.defenceUnit:SetAllWeaponsEnabled(false)		
	player.defenceUnit:SetCanBeKilled(false)
	player.defenceUnit:SetDoNotTarget(true)
	player.defenceUnit:SetMaxHealth(1)
	player.defenceUnit:SetHealth(nil,1)
	player.defenceUnit:SetRegenRate(1)
		
	--upgrade selection unit
	respawnUpgrader(player,positionX+bt.upgrader[1],positionY+bt.upgrader[2])
	
	--resource storage unit, needed to have resource storage for stats like money and zone control
	--spawn resource storage units here
	
	updateInfo(player)
end

--update player info
updateInfo = function(player)	
	if player.infoUnit != false then
		player.infoUnit:SetCustomName('Money ' .. username(player) .. ": " .. player.money)
	end
	if player.attackUnit != false then
		player.attackUnit:SetCustomName('Attack level: ' .. player.attack ..', upgrade for: ' .. (player.attack + 1)*50)
	end
	if player.defenceUnit != false then
		player.defenceUnit:SetCustomName('Defence level: ' .. player.defence ..', upgrade for: ' .. (player.defence + 1)*50)
	end
end






--~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~MISCALENIOUS FUNCTIONS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


--SPECIAL ARMY INITIALISATION: NO COMMANDER, NEUTRAL ALWAYS ENEMY
function InitializeArmies()
    local tblGroups = {}
    local tblArmy = ListArmies()
	
	--determine the players depending on which slots are occupied
	for index,army in tblArmy do
		players[armyTable[army]] = playersDATA[armyTable[army]]
	end
	players['neutral'].zonecontrol = 21 - playerCount() + 1
	
    for iArmy, strArmy in pairs(tblArmy) do
        local tblData = Scenario.Armies[strArmy]

        tblGroups[ strArmy ] = {}

        if tblData then


            for iEnemy, strEnemy in pairs(tblArmy) do
				
				--civilian ARMY_9 is always an enemey
                if iArmy != iEnemy and strArmy != 'NEUTRAL_CIVILIAN' and strEnemy != 'NEUTRAL_CIVILIAN' then
					SetAlliance( iArmy, iEnemy, 'Enemy')
                elseif strArmy == 'NEUTRAL_CIVILIAN' or strEnemy == 'NEUTRAL_CIVILIAN' then
                    SetAlliance( iArmy, iEnemy, 'Neutral')
                end
            end


        end

    end

    return tblGroups
end

